from Dictionary.TxtWord import TxtWord
from Dictionary.Word import Word

from MorphologicalAnalysis.InflectionalGroup import InflectionalGroup
from MorphologicalAnalysis.MorphologicalParse import MorphologicalParse
from MorphologicalAnalysis.State import State


class FsmParse(MorphologicalParse):
    __suffixList: list
    __formList: list
    __transitionList: list
    __withList: list
    __initialPos: str
    __pos: str
    __form: str
    __verbAgreement: str
    __possesiveAgreement: str

    """
    Another constructor of FsmParse class which takes a TxtWord root and a State as inputs.
    First, initializes root variable with this TxtWord. It also initializes form with root's name, pos and
    initialPos with given State's POS, creates 4 new list suffixList, formList, transitionList
    and withList and adds given State to suffixList, form to formList.

    PARAMETERS
    ----------
    root : TxtWord
        TxtWord input.
    startState : State
        State input.
    """

    def __init__(self, root, startState=None):
        if isinstance(root, Word):
            self.root = root
        elif isinstance(root, int):
            self.root = TxtWord(str(root))
            self.root.addFlag("IS_SAYI")
        elif isinstance(root, float):
            self.root = TxtWord(str(root))
            self.root.addFlag("IS_SAYI")
        elif isinstance(root, str):
            self.root = TxtWord(root)
        if startState is not None:
            self.__form = self.root.getName()
            self.__pos = startState.getPos()
            self.__initialPos = startState.getPos()
            self.__suffixList = []
            self.__suffixList.append(startState)
            self.__formList = []
            self.__formList.append(self.__form)
            self.__transitionList = []
            self.__withList = []
        self.__verbAgreement = None
        self.__possesiveAgreement = None

    def __eq__(self, other):
        return self.transitionList() == other.transitionList()

    def __lt__(self, other):
        return self.transitionList() < other.transitionList()

    def __gt__(self, other):
        return self.transitionList() > other.transitionList()

    """
    The constructInflectionalGroups method initially calls the transitionList method and assigns the resulting str
    to the parse variable and creates a new list as iGs. If parse str contains a derivational boundary
    it adds the substring starting from the 0 to the index of derivational boundary to the iGs. If it does not contain a 
    DB, it directly adds parse to the iGs. Then, creates and initializes new list as inflectionalGroups and fills with
    the items of iGs.
    """

    def constructInflectionalGroups(self):
        parse = self.transitionList()
        iGs = []
        while "^DB+" in parse:
            iGs.append(parse[:parse.index("^DB+")])
            parse = parse[parse.index("^DB+") + 4:]
        iGs.append(parse)
        self.inflectionalGroups = []
        self.inflectionalGroups.append(InflectionalGroup(iGs[0][iGs[0].index("+") + 1:]))
        for i in range(1, len(iGs)):
            self.inflectionalGroups.append(InflectionalGroup(iGs[i]))

    """
    Getter for the verbAgreement variable.

    RETURNS
    -------
    str
        The verbAgreement variable.
    """

    def getVerbAgreement(self) -> str:
        return self.__verbAgreement

    """
    Getter for the possesiveAgreement variable.
    
    RETURNS
    -------
    str
        The possesiveAgreement variable.
    """

    def getPossesiveAgreement(self) -> str:
        return self.__possesiveAgreement

    """
    The setAgreement method takes a str transitionName as an input and if it is one of the A1SG, A2SG, A3SG,
    A1PL, A2PL or A3PL it assigns transitionName input to the verbAgreement variable. Or if it is ine of the PNON, P1SG, 
    P2SG,P3SG, P1PL, P2PL or P3PL it assigns transitionName input to the possesiveAgreement variable.

    PARAMETERS
    ----------
    transitionName : str
        String input.
    """

    def setAgreement(self, transitionName: str):
        if transitionName == "A1SG" or transitionName == "A2SG" or transitionName == "A3SG" or transitionName == "A1PL"\
                or transitionName == "A2PL" or transitionName == "A3PL":
            self.__verbAgreement = transitionName
        if transitionName == "PNON" or transitionName == "P1SG" or transitionName == "P2SG" or transitionName == "P3SG"\
                or transitionName == "P1PL" or transitionName == "P2PL" or transitionName == "P3PL":
            self.__possesiveAgreement = transitionName

    """
    The getLastLemmaWithTag method takes a String input pos as an input. If given pos is an initial pos then it assigns
    root to the lemma, and assign null otherwise. Then, it loops i times where i ranges from 1 to size of the formList,
    if the item at i-1 of transitionList is not null and contains a derivational boundary with pos but not with ZERO,
    it assigns the ith item of formList to lemma.

    PARAMETERS
    ----------
    pos : str
        String input.
        
    RETURNS
    -------
    str
        String output lemma.
    """

    def getLastLemmaWithTag(self, pos: str) -> str:
        if self.__initialPos is not None and self.__initialPos == pos:
            lemma = self.root.getName()
        else:
            lemma = None
        for i in range(1, len(self.__formList)):
            if self.__transitionList[i - 1] is not None and ("^DB+" + pos) in self.__transitionList[i - 1] and \
                    ("^DB+" + pos + "+ZERO") in self.__transitionList[i - 1]:
                lemma = self.__formList[i]
        return lemma

    """
    The getLastLemma method initially assigns root as lemma. Then, it loops i times where i ranges from 1 to size of 
    the formList, if the item at i-1 of transitionList is not null and contains a derivational boundary, it assigns the 
    ith item of formList to lemma.

    RETURNS
    -------
    str
        String output lemma.
    """

    def getLastLemma(self) -> str:
        lemma = self.root.getName()
        for i in range(1, len(self.__formList)):
            if self.__transitionList[i - 1] is not None and "^DB+" in self.__transitionList[i - 1]:
                lemma = self.__formList[i]
        return lemma

    """
    The addSuffix method takes 5 different inputs; State suffix, str form, transition, with and toPos.
    If the pos of given input suffix is not None, it then assigns it to the pos variable. If the pos of the given suffix
    is None but given toPos is not None than it assigns toPos to pos variable. At the end, it adds suffix to the 
    suffixList, form to the formList, transition to the transitionList and if given with is not 0, it is also added to 
    withList.

    PARAMETERS
    ----------
    suffix : State    
        State input.
    form : str
        String input.
    transition : str
        String input.
    withName : str
        String input.
    toPos : str
        String input.
    """

    def addSuffix(self, suffix: State, form: str, transition: str, withName: str, toPos: str):
        if suffix.getPos() is not None:
            self.__pos = suffix.getPos()
        else:
            if toPos is not None:
                self.__pos = toPos
        self.__suffixList.append(suffix)
        self.__formList.append(form)
        self.__transitionList.append(transition)
        if withName != "0":
            self.__withList.append(withName)
        self.__form = form

    """
    Getter for the form variable.

    RETURNS
    -------
    str
        The form variable.
    """

    def getSurfaceForm(self) -> str:
        return self.__form

    """
    The getStartState method returns the first item of suffixList list.

    RETURNS
    -------
    State
        The first item of suffixList list.
    """

    def getStartState(self) -> State:
        return self.__suffixList[0]

    """
    Getter for the pos variable.

    RETURNS
    -------
    str
        The pos variable.
    """

    def getFinalPos(self) -> str:
        return self.__pos

    """
    Getter for the initialPos variable.

    RETURNS
    -------
    str
        The initialPos variable.
    """

    def getInitialPos(self) -> str:
        return self.__initialPos

    """
    The setForm method takes a str name as an input and assigns it to the form variable, then it removes the first item
    of formList list and adds the given name to the formList.

    PARAMETERS
    ----------
    name : str
        String input to set form.
    """

    def setForm(self, name: str):
        self.__form = name
        self.__formList.pop(0)
        self.__formList.append(name)

    """
    The getFinalSuffix method returns the last item of suffixList list.

    RETURNS
    -------
    State
        The last item of suffixList list.
    """

    def getFinalSuffix(self) -> State:
        return self.__suffixList[len(self.__suffixList) - 1]

    """
    The headerTransition method gets the first item of formList and checks for cases;

    If it is <DOC>, it returns "<DOC>+BDTAG" which indicates the beginning of a document.
    If it is </DOC>, it returns "</DOC>+EDTAG" which indicates the ending of a document.
    If it is <TITLE>, it returns "<TITLE>+BTTAG" which indicates the beginning of a title.
    If it is </TITLE>, it returns "</TITLE>+ETTAG" which indicates the ending of a title.
    If it is "<S>", it returns "<S>+BSTAG" which indicates the beginning of a sentence.
    If it is "</S>, it returns "</S>+ESTAG" which indicates the ending of a sentence.

    RETURNS
    -------
    str
        Corresponding tags of the headers and an empty {@link String} if any case does not match.
    """

    def headerTransition(self) -> str:
        if self.__formList[0] == "<DOC>":
            return "<DOC>+BDTAG"
        elif self.__formList[0] == "</DOC>":
            return "</DOC>+EDTAG"
        elif self.__formList[0] == "<TITLE>":
            return "<TITLE>+BTTAG"
        elif self.__formList[0] == "</TITLE>":
            return "</TITLE>+ETTAG"
        elif self.__formList[0] == "<S>":
            return "<S>+BSTAG"
        elif self.__formList[0] == "</S>":
            return "</S>+ESTAG"
        else:
            return ""

    """
    The pronounTransition method gets the first item of formList and checks for cases;

    If it is "kendi", it returns kendi+PRON+REFLEXP which indicates a reflexive pronoun.
    If it is one of the "hep, öbür, topu, öteki, kimse, hiçbiri, tümü, çoğu, hepsi, herkes, başkası, birçoğu, 
    birçokları, biri, birbirleri, birbiri, birkaçı, böylesi, diğeri, cümlesi, bazı, kimi", it returns
    +PRON+QUANTP which indicates a quantitative pronoun.
    If it is one of the "o, bu, şu" and if it is "o" it also checks the first item of suffixList and if it is a 
    PronounRoot(DEMONS), it returns +PRON+DEMONSP which indicates a demonstrative pronoun.
    If it is "ben", it returns +PRON+PERS+A1SG+PNON which indicates a 1st person singular agreement.
    If it is "sen", it returns +PRON+PERS+A2SG+PNON which indicates a 2nd person singular agreement.
    If it is "o" and the first item of suffixList, if it is a PronounRoot(PERS), it returns +PRON+PERS+A3SG+PNON which
    indicates a 3rd person singular agreement.
    If it is "biz", it returns +PRON+PERS+A1PL+PNON which indicates a 1st person plural agreement.
    If it is "siz", it returns +PRON+PERS+A2PL+PNON which indicates a 2nd person plural agreement.
    If it is "onlar" and the first item of suffixList, if it is a PronounRoot(PERS), it returns o+PRON+PERS+A3PL+PNON 
    which indicates a 3rd person plural agreement.
    If it is one of the "nere, ne, kim, hangi", it returns +PRON+QUESP which indicates a question pronoun.

    RETURNS
    -------
    str
        Corresponding transitions of pronouns and an empty str if any case does not match.
    """

    def pronounTransition(self) -> str:
        if self.__formList[0] == "kendi":
            return "kendi+PRON+REFLEXP"
        elif self.__formList[0] == "hep" or self.__formList[0] == "öbür" or self.__formList[0] == "topu" \
                or self.__formList[0] == "öteki" or self.__formList[0] == "kimse" or self.__formList[0] == "hiçbiri" \
                or self.__formList[0] == "tümü" or self.__formList[0] == "çoğu" or self.__formList[0] == "hepsi" \
                or self.__formList[0] == "herkes" or self.__formList[0] == "başkası" or self.__formList[0] == "birçoğu"\
                or self.__formList[0] == "birçokları" or self.__formList[0] == "birbiri" \
                or self.__formList[0] == "birbirleri" or self.__formList[0] == "biri" \
                or self.__formList[0] == "birkaçı" or self.__formList[0] == "böylesi" or self.__formList[0] == "diğeri"\
                or self.__formList[0] == "cümlesi" or self.__formList[0] == "bazı" or self.__formList[0] == "kimi":
            return self.__formList[0] + "+PRON+QUANTP"
        elif (self.__formList[0] == "o" and self.__suffixList[0].getName() == "PronounRoot(DEMONS)") \
                or self.__formList[0] == "bu" or self.__formList[0] == "şu":
            return self.__formList[0] + "+PRON+DEMONSP"
        elif self.__formList[0] == "ben":
            return self.__formList[0] + "+PRON+PERS+A1SG+PNON"
        elif self.__formList[0] == "sen":
            return self.__formList[0] + "+PRON+PERS+A2SG+PNON"
        elif self.__formList[0] == "o" and self.__suffixList[0].getName() == "PronounRoot(PERS)":
            return self.__formList[0] + "+PRON+PERS+A3SG+PNON"
        elif self.__formList[0] == "biz":
            return self.__formList[0] + "+PRON+PERS+A1PL+PNON"
        elif self.__formList[0] == "siz":
            return self.__formList[0] + "+PRON+PERS+A2PL+PNON"
        elif self.__formList[0] == "onlar":
            return "o+PRON+PERS+A3PL+PNON"
        elif self.__formList[0] == "nere" or self.__formList[0] == "ne" or self.__formList[0] == "kaçı" \
                or self.__formList[0] == "kim" or self.__formList[0] == "hangi":
            return self.__formList[0] + "+PRON+QUESP"
        else:
            return ""

    """
    The transitionList method first creates an empty {@link String} result, then gets the first item of suffixList and 
    checks for cases; If it is one of the "NominalRoot, NominalRootNoPossesive, CompoundNounRoot, NominalRootPlural", 
    it assigns concatenation of first item of formList and +NOUN to the result String.
    Ex : Birincilik

    If it is one of the "VerbalRoot, PassiveHn", it assigns concatenation of first item of formList and +VERB to the 
    result String.
    Ex : Başkalaştı

    If it is "CardinalRoot", it assigns concatenation of first item of formList and +NUM+CARD to the result String.
    Ex : Onuncu

    If it is "FractionRoot", it assigns concatenation of first item of formList and NUM+FRACTION to the result String.
    Ex : 1/2

    If it is "TimeRoot", it assigns concatenation of first item of formList and +TIME to the result String.
    Ex : 14:28

    If it is "RealRoot", it assigns concatenation of first item of formList and +NUM+REAL to the result String.
    Ex : 1.2

    If it is "Punctuation", it assigns concatenation of first item of formList and +PUNC to the result String.
    Ex : ,

    If it is "Hashtag", it assigns concatenation of first item of formList and +HASHTAG to the result String.
    Ex : #

    If it is "DateRoot", it assigns concatenation of first item of formList and +DATE to the result String.
    Ex : 11/06/2018

    If it is "RangeRoot", it assigns concatenation of first item of formList and +RANGE to the result String.
    Ex : 3-5

    If it is "Email", it assigns concatenation of first item of formList and +EMAIL to the result String.
    Ex : abc@

    If it is "PercentRoot", it assigns concatenation of first item of formList and +PERCENT to the result String.
    Ex : %12.5

    If it is "DeterminerRoot", it assigns concatenation of first item of formList and +DET to the result String.
    Ex : Birtakım

    If it is "ConjunctionRoot", it assigns concatenation of first item of formList and +CONJ to the result String.
    Ex : Ama

    If it is "AdverbRoot", it assigns concatenation of first item of formList and +ADV to the result String.
    Ex : Acilen

    If it is "ProperRoot", it assigns concatenation of first item of formList and +NOUN+PROP to the result String.
    Ex : Ahmet

    If it is "HeaderRoot", it assigns the result of the headerTransition method to the result String.
    Ex : &lt;DOC&gt;

    If it is "InterjectionRoot", it assigns concatenation of first item of formList and +INTERJ to the result String.
    Ex : Hey

    If it is "DuplicateRoot", it assigns concatenation of first item of formList and +DUP to the result String.
    Ex : Allak

    If it is "QuestionRoot", it assigns concatenation of first item of formList and +QUES to the result String.
    Ex : Mı

    If it is "PostP", and the first item of formList is one of the "karşı, ilişkin, göre, kadar, ait, yönelik, rağmen, 
    değin, dek, doğru, karşın, dair, atfen, binaen, hitaben, istinaden, mahsuben, mukabil, nazaran", it assigns 
    concatenation of first item of formList and +POSTP+PCDAT to the result String.
    Ex : İlişkin

    If it is "PostP", and the first item of formList is one of the "sonra, önce, beri, fazla, dolayı, itibaren, başka,
    çok, evvel, ötürü, yana, öte, aşağı, yukarı, dışarı, az, gayrı", it assigns concatenation of first
    item of formList and +POSTP+PCABL to the result String.
    Ex : Başka

    If it is "PostP", and the first item of formList is "yanısıra", it assigns concatenation of first
    item of formList and +POSTP+PCGEN to the result String.
    Ex : Yanısıra

    If it is "PostP", and the first item of formList is one of the "birlikte, beraber", it assigns concatenation of 
    first item of formList and +PPOSTP+PCINS to the result String.
    Ex : Birlikte

    If it is "PostP", and the first item of formList is one of the "aşkın, takiben", it assigns concatenation of first
    item of formList and +POSTP+PCACC to the result String.
    Ex : Takiben

    If it is "PostP", it assigns concatenation of first item of formList and +POSTP+PCNOM to the result String.

    If it is "PronounRoot", it assigns result of the pronounTransition method to the result String.
    Ex : Ben

    If it is "OrdinalRoot", it assigns concatenation of first item of formList and +NUM+ORD to the result String.
    Ex : Altıncı

    If it starts with "Adjective", it assigns concatenation of first item of formList and +ADJ to the result String.
    Ex : Güzel

    At the end, it loops through the formList and concatenates each item with result {@link String}.

    RETURNS
    -------
    str
        str result accumulated with items of formList.
    """

    def transitionList(self) -> str:
        result = ""
        if self.__suffixList[0].getName() == "NominalRoot" \
                or self.__suffixList[0].getName() == "NominalRootNoPossesive" \
                or self.__suffixList[0].getName() == "CompoundNounRoot" \
                or self.__suffixList[0].getName() == "NominalRootPlural":
            result = self.__formList[0] + "+NOUN"
        elif self.__suffixList[0].getName().startswith("VerbalRoot") or self.__suffixList[0].getName() == "PassiveHn":
            result = self.__formList[0] + "+VERB"
        elif self.__suffixList[0].getName() == "CardinalRoot":
            result = self.__formList[0] + "+NUM+CARD"
        elif self.__suffixList[0].getName() == "FractionRoot":
            result = self.__formList[0] + "+NUM+FRACTION"
        elif self.__suffixList[0].getName() == "TimeRoot":
            result = self.__formList[0] + "+TIME"
        elif self.__suffixList[0].getName() == "RealRoot":
            result = self.__formList[0] + "+NUM+REAL"
        elif self.__suffixList[0].getName() == "Punctuation":
            result = self.__formList[0] + "+PUNC"
        elif self.__suffixList[0].getName() == "Hashtag":
            result = self.__formList[0] + "+HASHTAG"
        elif self.__suffixList[0].getName() == "DateRoot":
            result = self.__formList[0] + "+DATE"
        elif self.__suffixList[0].getName() == "RangeRoot":
            result = self.__formList[0] + "+RANGE"
        elif self.__suffixList[0].getName() == "Email":
            result = self.__formList[0] + "+EMAIL"
        elif self.__suffixList[0].getName() == "PercentRoot":
            result = self.__formList[0] + "+PERCENT"
        elif self.__suffixList[0].getName() == "DeterminerRoot":
            result = self.__formList[0] + "+DET"
        elif self.__suffixList[0].getName() == "ConjunctionRoot":
            result = self.__formList[0] + "+CONJ"
        elif self.__suffixList[0].getName() == "AdverbRoot":
            result = self.__formList[0] + "+ADV"
        elif self.__suffixList[0].getName() == "ProperRoot":
            result = self.__formList[0] + "+NOUN+PROP"
        elif self.__suffixList[0].getName() == "HeaderRoot":
            result = self.headerTransition()
        elif self.__suffixList[0].getName() == "InterjectionRoot":
            result = self.__formList[0] + "+INTERJ"
        elif self.__suffixList[0].getName() == "DuplicateRoot":
            result = self.__formList[0] + "+DUP"
        elif self.__suffixList[0].getName() == "QuestionRoot":
            result = self.__formList[0] + "+QUES"
        elif self.__suffixList[0].getName() == "PostP":
            if self.__formList[0] == "karşı" or self.__formList[0] == "ilişkin" or self.__formList[0] == "göre" \
                    or self.__formList[0] == "kadar" or self.__formList[0] == "ait" or self.__formList[0] == "yönelik" \
                    or self.__formList[0] == "rağmen" or self.__formList[0] == "değin" or self.__formList[0] == "dek" \
                    or self.__formList[0] == "doğru" or self.__formList[0] == "karşın" or self.__formList[0] == "dair" \
                    or self.__formList[0] == "atfen" or self.__formList[0] == "binaen" or self.__formList[
                0] == "hitaben" \
                    or self.__formList[0] == "istinaden" or self.__formList[0] == "mahsuben" \
                    or self.__formList[0] == "mukabil" or self.__formList[0] == "nazaran":
                result = self.__formList[0] + "+POSTP+PCDAT"
            elif self.__formList[0] == "sonra" or self.__formList[0] == "önce" or self.__formList[0] == "beri" \
                    or self.__formList[0] == "fazla" or self.__formList[0] == "dolayı" or self.__formList[
                0] == "itibaren" \
                    or self.__formList[0] == "başka" or self.__formList[0] == "çok" or self.__formList[0] == "evvel" \
                    or self.__formList[0] == "ötürü" or self.__formList[0] == "yana" or self.__formList[0] == "öte" \
                    or self.__formList[0] == "aşağı" or self.__formList[0] == "yukarı" \
                    or self.__formList[0] == "dışarı" or self.__formList[0] == "az" or self.__formList[0] == "gayrı":
                result = self.__formList[0] + "+POSTP+PCABL"
            elif self.__formList[0] == "yanısıra":
                result = self.__formList[0] + "+POSTP+PCGEN"
            elif self.__formList[0] == "birlikte" or self.__formList[0] == "beraber":
                result = self.__formList[0] + "+POSTP+PCINS"
            elif self.__formList[0] == "aşkın" or self.__formList[0] == "takiben":
                result = self.__formList[0] + "+POSTP+PCACC"
            else:
                result = self.__formList[0] + "+POSTP+PCNOM"
        elif self.__suffixList[0].getName().startswith("PronounRoot"):
            result = self.pronounTransition()
        elif self.__suffixList[0].getName() == "OrdinalRoot":
            result = self.__formList[0] + "+NUM+ORD"
        elif self.__suffixList[0].getName().startswith("Adjective"):
            result = self.__formList[0] + "+ADJ"
        for transition in self.__transitionList:
            if transition is not None:
                if not transition.startswith("^"):
                    result = result + "+" + transition
                else:
                    result = result + transition
        return result

    """
    The suffixList method gets the first items of suffixList and formList and concatenates them with parenthesis and
    assigns a String result. Then, loops through the formList and it the current ith item is not equal to previous
    item it accumulates ith items of formList and suffixList to the result str.

    RETURNS
    -------
    str
        result str accumulated with the items of formList and suffixList.
    """

    def suffixList(self) -> str:
        result = self.__suffixList[0].getName() + "(" + self.__formList[0] + ")"
        for i in range(1, len(self.__formList)):
            if self.__formList[i] != self.__formList[i - 1]:
                result = result + self.__suffixList[i].getName() + "(" + self.__formList[i] + ")"
        return result

    """
    The withList method gets the root as a result str then loops through the withList and concatenates each item
    with result str.
    
    RETURNS
    -------
    str
        result str accumulated with items of withList.
    """

    def withList(self) -> str:
        result = self.root.getName()
        for aWith in self.__withList:
            result = result + "+" + aWith
        return result

    """
    The overridden str method which returns transitionList method.

    RETURNS
    -------
    str
        Returns transitionList method.
    """

    def __str__(self) -> str:
        return self.transitionList()



class State:

    __startState: bool
    __endState: bool
    __name: str
    __pos: str

    """
    Second constructor of the State class which takes 4 parameters as input; String name, boolean startState,
    boolean endState, and String pos and initializes the private variables of the class.

    PARAMETERS
    ----------
    name : str
        String input.
    startState : bool
        boolean input.
    endState : bool  
        boolean input.
    pos : str
        String input.
    """
    def __init__(self, name: str, startState: bool, endState: bool, pos=None):
        self.__endState = endState
        self.__startState = startState
        self.__name = name
        self.__pos = pos

    """
    Overridden __str__ method which returns the name.

    RETURNS
    -------
    str
        String name.
    """
    def __str__(self) -> str:
        return self.__name

    """
    Getter for the name.

    RETURNS
    -------
    str
        String name.
    """
    def getName(self) -> str:
        return self.__name

    """
    Getter for the pos.

    RETURNS
    -------
    str
        String pos.
    """
    def getPos(self) -> str:
        return self.__pos

    """
    The isEndState method returns endState's value.

    RETURNS
    -------
    bool
        boolean endState.
    """
    def isEndState(self) -> bool:
        return self.__endState



import xml.etree.ElementTree

from MorphologicalAnalysis.State import State
from MorphologicalAnalysis.Transition import Transition


class FiniteStateMachine:

    __states: list
    __transitions: dict

    """
    Constructor reads the finite state machine in the given input file. It has a NodeList which holds the states
    of the nodes and there are 4 different type of nodes; stateNode, root Node, transitionNode and withNode.
    Also there are two states; state that a node currently in and state that a node will be in.

    XMLParser is used to parse the given file. Firstly it gets the document to parse, then gets its elements by the
    tag names. For instance, it gets states by the tag name 'state' and puts them into an ArrayList called stateList.
    Secondly, it traverses this stateList and gets each Node's attributes. There are three attributes; name, start,
    and end which will be named as states. If a node is in a startState it is tagged as 'yes', otherwise 'no'.
    Also, if a node is in a startState, additional attribute will be fetched; originalPos that represents its original
    part of speech.

    At the last step, by starting rootNode's first child, it gets all the transitionNodes and next states called toState
    then continue with the nextSiblings. Also, if there is no possible toState, it prints this case and the causative 
    states.

    PARAMETERS
    ----------
    fileName : str
        the resource file to read the finite state machine. Only files in resources folder are supported.
    """
    def __init__(self, fileName: str):
        self.__transitions = {}
        self.__states = []
        root = xml.etree.ElementTree.parse(fileName).getroot()
        for stateNode in root:
            stateName = stateNode.attrib["name"]
            startState = stateNode.attrib["start"] == "yes"
            endState = stateNode.attrib["end"] == "yes"
            if startState:
                originalPos = stateNode.attrib["originalpos"]
                self.__states.append(State(stateName, True, endState, originalPos))
            else:
                self.__states.append(State(stateName, False, endState))
        for stateNode in root:
            if "name" in stateNode.attrib:
                stateName = stateNode.attrib["name"]
                state = self.getState(stateName)
                for transitionNode in stateNode:
                    stateName = transitionNode.attrib["name"]
                    if "transitionname" in transitionNode.attrib:
                        withName = transitionNode.attrib["transitionname"]
                    else:
                        withName = None
                    if "topos" in transitionNode.attrib:
                        rootToPos = transitionNode.attrib["topos"]
                    else:
                        rootToPos = None
                    toState = self.getState(stateName)
                    if toState is not None:
                        for withNode in transitionNode:
                            if "name" in withNode.attrib:
                                withName = withNode.attrib["name"]
                                if "topos" in withNode.attrib:
                                    toPos = withNode.attrib["topos"]
                                else:
                                    toPos = None
                            else:
                                toPos = None
                            if toPos is None:
                                if rootToPos is None:
                                    self.addTransition(state, toState, withNode.text, withName)
                                else:
                                    self.addTransition(state, toState, withNode.text, withName, rootToPos)
                            else:
                                self.addTransition(state, toState, withNode.text, withName, toPos)

    """
    The isValidTransition loops through states ArrayList and checks transitions between states. If the actual transition
    equals to the given transition input, method returns true otherwise returns false.

    PARAMETERS
    ----------
    transition : str
        is used to compare with the actual transition of a state.
        
    RETURNS
    -------
    bool
        True when the actual transition equals to the transition input, false otherwise.
    """
    def isValidTransition(self, transition: str) -> bool:
        for state in self.__transitions.keys():
            for transition1 in self.__transitions[state]:
                if transition1.__str__() is not None and transition1.__str__() == transition:
                    return True
        return False

    """
    The getState method is used to loop through the states list and return the state whose name equal
    to the given input name.

    PARAMETERS
    ----------
    name : str
        is used to compare with the state's actual name.
        
    RETURNS
    -------
    State
        State if found any, None otherwise.
    """
    def getState(self, name: str) -> State:
        for state in self.__states:
            if state.getName() == name:
                return state
        return None

    """
    Another addTransition method which takes additional argument; toPos and. It creates a new Transition
    with given input parameters and adds the transition to transitions list.

    PARAMETERS
    ----------
    toState : State 
        State type input indicating the next state.
    _with : str    
        String input indicating with what the transition will be made.
    withName : str
        String input.
    toPos : str   
        String input.
    """
    def addTransition(self, fromState: State, toState: State, _with: str, withName: str, toPos=None):
        newTransition = Transition(_with, toState, withName, toPos)
        if fromState in self.__transitions:
            transitionList = self.__transitions[fromState]
        else:
            transitionList = []
        transitionList.append(newTransition)
        self.__transitions[fromState] = transitionList

    """
    The getTransitions method returns the transitions at the given state.

    PARAMETERS
    ----------
    state : State
        State input.
        
    RETURNS
    -------
    list
        Transitions at given state.
    """
    def getTransitions(self, state: State) -> list:
        if state in self.__transitions:
            return self.__transitions[state]
        else:
            return []



import copy
import re

from Corpus.Sentence import Sentence
from DataStructure.Cache.LRUCache import LRUCache
from Dictionary.Trie.Trie import Trie
from Dictionary.TxtDictionary import TxtDictionary
from Dictionary.TxtWord import TxtWord
from Dictionary.Word import Word

from MorphologicalAnalysis.FiniteStateMachine import FiniteStateMachine
from MorphologicalAnalysis.FsmParse import FsmParse
from MorphologicalAnalysis.FsmParseList import FsmParseList
from MorphologicalAnalysis.MetamorphicParse import MetamorphicParse
from MorphologicalAnalysis.MorphologicalParse import MorphologicalParse
from MorphologicalAnalysis.MorphologicalTag import MorphologicalTag
from MorphologicalAnalysis.State import State
from MorphologicalAnalysis.Transition import Transition


class FsmMorphologicalAnalyzer:
    __dictionaryTrie: Trie
    __finiteStateMachine: FiniteStateMachine
    __dictionary: TxtDictionary
    __cache: LRUCache
    __mostUsedPatterns = {}

    MAX_DISTANCE = 2

    """
    Constructor of FsmMorphologicalAnalyzer class. It generates a new TxtDictionary type dictionary from
    given input dictionary file name and by using turkish_finite_state_machine.xml file.

    PARAMETERS
    ----------
    fileName : str          
        the file to read the finite state machine.  
    cacheSize : int
        the size of the LRUCache.
    dictionaryFileName : str
        the file to read the dictionary.
    misspelledFileName: str
        the file to read the misspelled file name.
    """

    def __init__(self, dictionaryFileName=None, misspelledFileName=None, fileName="turkish_finite_state_machine.xml",
                 cacheSize=10000000):
        if dictionaryFileName is None:
            self.__dictionary = TxtDictionary()
        else:
            self.__dictionary = TxtDictionary(dictionaryFileName, misspelledFileName)
        self.__finiteStateMachine = FiniteStateMachine(fileName)
        self.__dictionaryTrie = self.__dictionary.prepareTrie()
        self.__cache = LRUCache(cacheSize)

    """
    The getPossibleWords method takes MorphologicalParse and MetamorphicParse as input.
    First it determines whether the given morphologicalParse is the root verb and whether it contains a verb tag.
    Then it creates new transition with -mak and creates a new set result.

    It takes the given MetamorphicParse input as currentWord and if there is a compound word starting with the
    currentWord, it gets this compoundWord from dictionaryTrie. If there is a compoundWord and the difference of the
    currentWord and compundWords is less than 3 than compoundWord is added to the result, otherwise currentWord is added

    Then it gets the root from parse input as a currentRoot. If it is not null, and morphologicalParse input is verb,
    it directly adds the verb to result after making transition to currentRoot with currentWord String. Else, it creates 
    a new transition with -lar and make this transition then adds to the result.

    PARAMETERS
    ----------
    morphologicalParse : MorphologicalParse 
        MorphologicalParse type input.
    metamorphicParse : MetamorphicParse
        MetamorphicParse type input.
        
    RETURNS
    -------
    set
        set result.
    """

    def getPossibleWords(self, morphologicalParse: MorphologicalParse, metamorphicParse: MetamorphicParse) -> set:
        isRootVerb = morphologicalParse.getRootPos() == "VERB"
        containsVerb = morphologicalParse.containsTag(MorphologicalTag.VERB)
        verbTransition = Transition("mAk")
        result = set()
        if metamorphicParse is None or metamorphicParse.getWord() is None:
            return result
        currentWord = metamorphicParse.getWord().getName()
        pluralIndex = -1
        compoundWord = self.__dictionaryTrie.getCompundWordStartingWith(currentWord)
        if not isRootVerb:
            if compoundWord is not None and len(compoundWord.getName()) - len(currentWord) < 3:
                result.add(compoundWord.getName())
            result.add(currentWord)
        currentRoot = self.__dictionary.getWord(metamorphicParse.getWord().getName())
        if currentRoot is None and compoundWord is not None:
            currentRoot = compoundWord
        if currentRoot is not None:
            if isRootVerb:
                verbWord = verbTransition.makeTransitionNoStartState(currentRoot, currentWord)
                result.add(verbWord)
            pluralWord = None
            for i in range(1, metamorphicParse.size()):
                transition = Transition(metamorphicParse.getMetaMorpheme(i))
                if metamorphicParse.getMetaMorpheme(i) == "lAr":
                    pluralWord = currentWord
                    pluralIndex = i + 1
                currentWord = transition.makeTransitionNoStartState(currentRoot, currentWord)
                result.add(currentWord)
                if containsVerb:
                    verbWord = verbTransition.makeTransitionNoStartState(currentRoot, currentWord)
                    result.add(verbWord)
            if pluralWord is not None:
                currentWord = pluralWord
                for i in range(pluralIndex, metamorphicParse.size()):
                    transition = Transition(metamorphicParse.getMetaMorpheme(i))
                    currentWord = transition.makeTransitionNoStartState(currentRoot, currentWord)
                    result.add(currentWord)
                    if containsVerb:
                        verbWord = verbTransition.makeTransitionNoStartState(currentRoot, currentWord)
                        result.add(verbWord)
        return result

    """
    The getDictionary method is used to get TxtDictionary.

    RETURNS
    -------
    TxtDictionary
        TxtDictionary type dictionary.
    """

    def getDictionary(self) -> TxtDictionary:
        return self.__dictionary

    """
    The getFiniteStateMachine method is used to get FiniteStateMachine.

    RETURNS
    -------
    FiniteStateMachine
        FiniteStateMachine type finiteStateMachine.
    """

    def getFiniteStateMachine(self) -> FiniteStateMachine:
        return self.__finiteStateMachine

    """
    The isPossibleSubstring method first checks whether given short and long strings are equal to root word.
    Then, compares both short and long strings' chars till the last two chars of short string. In the presence of 
    mismatch, false is returned. On the other hand, it counts the distance between two strings until it becomes greater 
    than 2, which is the MAX_DISTANCE also finds the index of the last char.

    If the substring is a rootWord and equals to 'ben', which is a special case or root holds the 
    lastIdropsDuringSuffixation or lastIdropsDuringPassiveSuffixation conditions, then it returns true if distance is 
    not greater than MAX_DISTANCE.

    On the other hand, if the shortStrong ends with one of these chars 'e, a, p, ç, t, k' and it 's a rootWord with
    the conditions of rootSoftenDuringSuffixation, vowelEChangesToIDuringYSuffixation, 
    vowelAChangesToIDuringYSuffixation or endingKChangesIntoG then it returns true if the last index is not equal to 2 
    and distance is not greater than MAX_DISTANCE and false otherwise.

    PARAMETERS
    ----------
    shortString : str
        the possible substring.
    longString : str
        the long string to compare with substring.
    root : TxtWord
        the root of the long string.
        
    RETURNS
    -------
    bool
        True if given substring is the actual substring of the longString, false otherwise.
    """

    def isPossibleSubstring(self, shortString: str, longString: str, root: TxtWord) -> bool:
        rootWord = shortString == root.getName() or longString == root.getName()
        distance = 0
        last = 1
        for j in range(len(shortString)):
            if shortString[j] != longString[j]:
                if j < len(shortString) - 2:
                    return False
                last = len(shortString) - j
                distance = distance + 1
                if distance > self.MAX_DISTANCE:
                    break
        if rootWord and (root.getName() == "ben" or root.lastIdropsDuringSuffixation()
                         or root.lastIdropsDuringPassiveSuffixation()):
            return distance <= self.MAX_DISTANCE
        elif shortString.endswith("e") or shortString.endswith("a") or shortString.endswith("p") \
                or shortString.endswith("ç") or shortString.endswith("t") or shortString.endswith("k") \
                or (rootWord and (root.rootSoftenDuringSuffixation() or root.vowelEChangesToIDuringYSuffixation()
                                  or root.vowelAChangesToIDuringYSuffixation() or root.endingKChangesIntoG())):
            return last != 2 and distance <= self.MAX_DISTANCE - 1
        else:
            return distance <= self.MAX_DISTANCE - 2

    """
    The initializeParseList method initializes the given given fsm ArrayList with given root words by parsing them.

    It checks many conditions;
    isPlural; if root holds the condition then it gets the state with the name of NominalRootPlural, then
    creates a new parsing and adds this to the input fsmParse Arraylist.
    Ex : Açıktohumlular

    !isPlural and isPortmanteauEndingWithSI, if root holds the conditions then it gets the state with the
    name of NominalRootNoPossesive.
    Ex : Balarısı

    !isPlural and isPortmanteau, if root holds the conditions then it gets the state with the name of
    CompoundNounRoot.
    Ex : Aslanağızı

    !isPlural, !isPortmanteau and isHeader, if root holds the conditions then it gets the state with the
    name of HeaderRoot.
    Ex :  </title>

    !isPlural, !isPortmanteau and isInterjection, if root holds the conditions then it gets the state
    with the name of InterjectionRoot.
    Ex : Hey, Aa

    !isPlural, !isPortmanteau and isDuplicate, if root holds the conditions then it gets the state
    with the name of DuplicateRoot.

    !isPlural, !isPortmanteau and isNumeral, if root holds the conditions then it gets the state
    with the name of CardinalRoot.

    !isPlural, !isPortmanteau and isReal, if root holds the conditions then it gets the state
    with the name of RealRoot.

    !isPlural, !isPortmanteau and isFraction, if root holds the conditions then it gets the state
    with the name of FractionRoot.
    Ex : 1/2

    !isPlural, !isPortmanteau and isDate, if root holds the conditions then it gets the state
    with the name of DateRoot.
    Ex : 11/06/2018

    !isPlural, !isPortmanteau and isPercent, if root holds the conditions then it gets the state
    with the name of PercentRoot.
    Ex : %12.5

    !isPlural, !isPortmanteau and isRange, if root holds the conditions then it gets the state
    with the name of RangeRoot.
    Ex : 3-5

    !isPlural, !isPortmanteau and isTime, if root holds the conditions then it gets the state
    with the name of TimeRoot.
    Ex : 13:16:08

    !isPlural, !isPortmanteau and isOrdinal, if root holds the conditions then it gets the state
    with the name of OrdinalRoot.
    Ex : Altıncı

    !isPlural, !isPortmanteau, and isVerb if root holds the conditions then it gets the state
    with the name of VerbalRoot. Or isPassive, then it gets the state with the name of PassiveHn.
    Ex : Anla (!isPAssive)
    Ex : Çağrıl (isPassive)

    !isPlural, !isPortmanteau and isPronoun, if root holds the conditions then it gets the state
    with the name of PronounRoot. There are 6 different Pronoun state names, REFLEX, QUANT, QUANTPLURAL, DEMONS, PERS, 
    QUES.
    REFLEX = Reflexive Pronouns Ex : kendi
    QUANT = Quantitative Pronouns Ex : öbür, hep, kimse, hiçbiri, bazı, kimi, biri
    QUANTPLURAL = Quantitative Plural Pronouns Ex : tümü, çoğu, hepsi
    DEMONS = Demonstrative Pronouns Ex : o, bu, şu
    PERS = Personal Pronouns Ex : ben, sen, o, biz, siz, onlar
    QUES = Interrogatıve Pronouns Ex : nere, ne, kim, hangi

    !isPlural, !isPortmanteau and isAdjective, if root holds the conditions then it gets the state
    with the name of AdjectiveRoot.
    Ex : Absürt, Abes

    !isPlural, !isPortmanteau and isPureAdjective, if root holds the conditions then it gets the state
    with the name of Adjective.
    Ex : Geçmiş, Cam

    !isPlural, !isPortmanteau and isNominal, if root holds the conditions then it gets the state
    with the name of NominalRoot.
    Ex : Görüş

    !isPlural, !isPortmanteau and isProper, if root holds the conditions then it gets the state
    with the name of ProperRoot.
    Ex : Abdi

    !isPlural, !isPortmanteau and isQuestion, if root holds the conditions then it gets the state
    with the name of QuestionRoot.
    Ex : Mi, mü

    !isPlural, !isPortmanteau and isDeterminer, if root holds the conditions then it gets the state
    with the name of DeterminerRoot.
    Ex : Çok, bir

    !isPlural, !isPortmanteau and isConjunction, if root holds the conditions then it gets the state
    with the name of ConjunctionRoot.
    Ex : Ama , ancak

    !isPlural, !isPortmanteau and isPostP, if root holds the conditions then it gets the state
    with the name of PostP.
    Ex : Ait, dair

    !isPlural, !isPortmanteau and isAdverb, if root holds the conditions then it gets the state
    with the name of AdverbRoot.
    Ex : Acilen

    PARAMETERS
    ----------
    fsmParse : list
        list to initialize.
    root : TxtWord    
        word to check properties and add to fsmParse according to them.
    isProper : bool
        is used to check a word is proper or not.
    """

    def initializeParseList(self, fsmParse: list, root: TxtWord, isProper: bool):
        if root.isPlural():
            currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("NominalRootPlural"))
            fsmParse.append(currentFsmParse)
        elif root.isPortmanteauEndingWithSI():
            currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 2],
                                       self.__finiteStateMachine.getState("CompoundNounRoot"))
            fsmParse.append(currentFsmParse)
            currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("NominalRootNoPossesive"))
            fsmParse.append(currentFsmParse)
        elif root.isPortmanteau():
            if root.isPortmanteauFacedVowelEllipsis():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("NominalRootNoPossesive"))
                fsmParse.append(currentFsmParse)
                currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 2] +
                                           root.getName()[len(root.getName()) - 1] +
                                           root.getName()[len(root.getName()) - 2],
                                           self.__finiteStateMachine.getState("CompoundNounRoot"))
            elif root.isPortmanteauFacedSoftening():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("NominalRootNoPossesive"))
                fsmParse.append(currentFsmParse)
                if root.getName()[len(root.getName()) - 2] == "b":
                    currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 2] + "p",
                                               self.__finiteStateMachine.getState("CompoundNounRoot"))
                elif root.getName()[len(root.getName()) - 2] == "c":
                    currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 2] + "ç",
                                               self.__finiteStateMachine.getState("CompoundNounRoot"))
                elif root.getName()[len(root.getName()) - 2] == "d":
                    currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 2] + "t",
                                               self.__finiteStateMachine.getState("CompoundNounRoot"))
                elif root.getName()[len(root.getName()) - 2] == "ğ":
                    currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 2] + "k",
                                               self.__finiteStateMachine.getState("CompoundNounRoot"))
                else:
                    currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 1],
                                               self.__finiteStateMachine.getState("CompoundNounRoot"))
            else:
                currentFsmParse = FsmParse(root.getName()[:len(root.getName()) - 1],
                                           self.__finiteStateMachine.getState("CompoundNounRoot"))
            fsmParse.append(currentFsmParse)
        else:
            if root.isHeader():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("HeaderRoot"))
                fsmParse.append(currentFsmParse)
            if root.isInterjection():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("InterjectionRoot"))
                fsmParse.append(currentFsmParse)
            if root.isDuplicate():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("DuplicateRoot"))
                fsmParse.append(currentFsmParse)
            if root.isNumeral():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("CardinalRoot"))
                fsmParse.append(currentFsmParse)
            if root.isReal():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("RealRoot"))
                fsmParse.append(currentFsmParse)
            if root.isFraction():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("FractionRoot"))
                fsmParse.append(currentFsmParse)
            if root.isDate():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("DateRoot"))
                fsmParse.append(currentFsmParse)
            if root.isPercent():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PercentRoot"))
                fsmParse.append(currentFsmParse)
            if root.isRange():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("RangeRoot"))
                fsmParse.append(currentFsmParse)
            if root.isTime():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("TimeRoot"))
                fsmParse.append(currentFsmParse)
            if root.isOrdinal():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("OrdinalRoot"))
                fsmParse.append(currentFsmParse)
            if root.isVerb() or root.isPassive():
                if root.verbType() != "":
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("VerbalRoot("
                                                                                        + root.verbType() + ")"))
                elif not root.isPassive():
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("VerbalRoot"))
                else:
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PassiveHn"))
                fsmParse.append(currentFsmParse)
            if root.isPronoun():
                if root.getName() == "kendi":
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PronounRoot(REFLEX)"))
                    fsmParse.append(currentFsmParse)
                if root.getName() == "öbür" or root.getName() == "öteki" or root.getName() == "hep" or \
                        root.getName() == "kimse" or root.getName() == "diğeri" or root.getName() == "hiçbiri" or \
                        root.getName() == "böylesi" or root.getName() == "birbiri" or root.getName() == "birbirleri" or\
                        root.getName() == "biri" or root.getName() == "başkası" or root.getName() == "bazı" or \
                        root.getName() == "kimi":
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PronounRoot(QUANT)"))
                    fsmParse.append(currentFsmParse)
                if root.getName() == "tümü" or root.getName() == "topu" or root.getName() == "herkes" or \
                        root.getName() == "cümlesi" or root.getName() == "çoğu" or root.getName() == "birçoğu" or \
                        root.getName() == "birkaçı" or root.getName() == "birçokları" or root.getName() == "hepsi":
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PronounRoot(QUANTPLURAL)"))
                    fsmParse.append(currentFsmParse)
                if root.getName() == "o" or root.getName() == "bu" or root.getName() == "şu":
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PronounRoot(DEMONS)"))
                    fsmParse.append(currentFsmParse)
                if root.getName() == "ben" or root.getName() == "sen" or root.getName() == "o" or \
                        root.getName() == "biz" or root.getName() == "siz" or root.getName() == "onlar":
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PronounRoot(PERS)"))
                    fsmParse.append(currentFsmParse)
                if root.getName() == "nere" or root.getName() == "ne" or root.getName() == "kaçı" or \
                        root.getName() == "kim" or root.getName() == "hangi":
                    currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PronounRoot(QUES)"))
                    fsmParse.append(currentFsmParse)
            if root.isAdjective():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("AdjectiveRoot"))
                fsmParse.append(currentFsmParse)
            if root.isPureAdjective():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("Adjective"))
                fsmParse.append(currentFsmParse)
            if root.isNominal():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("NominalRoot"))
                fsmParse.append(currentFsmParse)
            if root.isAbbreviation():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("NominalRoot"))
                fsmParse.append(currentFsmParse)
            if root.isProperNoun() and isProper:
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("ProperRoot"))
                fsmParse.append(currentFsmParse)
            if root.isQuestion():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("QuestionRoot"))
                fsmParse.append(currentFsmParse)
            if root.isDeterminer():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("DeterminerRoot"))
                fsmParse.append(currentFsmParse)
            if root.isConjunction():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("ConjunctionRoot"))
                fsmParse.append(currentFsmParse)
            if root.isPostP():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("PostP"))
                fsmParse.append(currentFsmParse)
            if root.isAdverb():
                currentFsmParse = FsmParse(root, self.__finiteStateMachine.getState("AdverbRoot"))
                fsmParse.append(currentFsmParse)

    """
    The initializeParseListFromRoot method is used to create a list which consists of initial fsm parsings. First, 
    traverses this HashSet and uses each word as a root and calls initializeParseList method with this root and list.

    PARAMETERS
    ----------
    parseList : list
        list to initialize.
    root : TxtWord
        the root form to generate initial parse list.
    isProper : bool   
        is used to check a word is proper or not.
    """

    def initializeParseListFromRoot(self, parseList: list, root: TxtWord, isProper: bool):
        self.initializeParseList(parseList, root, isProper)
        if root.obeysAndNotObeysVowelHarmonyDuringAgglutination():
            newRoot = copy.deepcopy(root)
            newRoot.removeFlag("IS_UU")
            newRoot.removeFlag("IS_UUU")
            self.initializeParseList(parseList, newRoot, isProper)
        if root.rootSoftenAndNotSoftenDuringSuffixation():
            newRoot = copy.deepcopy(root)
            newRoot.removeFlag("IS_SD")
            newRoot.removeFlag("IS_SDD")
            self.initializeParseList(parseList, newRoot, isProper)
        if root.lastIDropsAndNotDropDuringSuffixation():
            newRoot = copy.deepcopy(root)
            newRoot.removeFlag("IS_UD")
            newRoot.removeFlag("IS_UDD")
            self.initializeParseList(parseList, newRoot, isProper)
        if root.duplicatesAndNotDuplicatesDuringSuffixation():
            newRoot = copy.deepcopy(root)
            newRoot.removeFlag("IS_ST")
            newRoot.removeFlag("IS_STT")
            self.initializeParseList(parseList, newRoot, isProper)
        if root.endingKChangesIntoG() and root.containsFlag("IS_OA"):
            newRoot = copy.deepcopy(root)
            newRoot.removeFlag("IS_OA")
            self.initializeParseList(parseList, newRoot, isProper)

    """
    The initializeParseListFromSurfaceForm method is used to create a list which consists of initial fsm parsings. 
    First, it calls getWordsWithPrefix methods by using input String surfaceForm and generates a set. Then, traverses
    this set and uses each word as a root and calls initializeParseListFromRoot method with this root and list.

    PARAMETERS
    ----------
    surfaceForm : str
        the String used to generate a HashSet of words.
    isProper : bool   
        is used to check a word is proper or not.
        
    RETURNS
    -------
    list
        initialFsmParse list.
    """

    def initializeParseListFromSurfaceForm(self, surfaceForm: str, isProper: bool) -> list:
        initialFsmParse = []
        if len(surfaceForm) == 0:
            return initialFsmParse
        words = self.__dictionaryTrie.getWordsWithPrefix(surfaceForm)
        for word in words:
            self.initializeParseListFromRoot(initialFsmParse, word, isProper)
        return initialFsmParse

    """
    The addNewParsesFromCurrentParseSurfaceForm method initially gets the final suffixes from input currentFsmParse 
    called as currentState, and by using the currentState information it gets the currentSurfaceForm. Then loops 
    through each currentState's transition. If the currentTransition is possible, it makes the transition.
    The addNewParsesFromCurrentParseMaxLength method initially gets the final suffixes from input currentFsmParse 
    called as currentState, and by using the currentState information it gets the new analysis. Then loops through each 
    currentState's transition. If the currentTransition is possible, it makes the transition.

    PARAMETERS
    ----------
    currentFsmParse : FsmParse
        FsmParse type input.
    fsmParse : list
        List of FsmParse.
    maxLength : int    
        Maximum length of the parse.
    surfaceForm : str    
        String to use during transition.
    root : TxtWord           
        TxtWord used to make transition.
    """

    def addNewParsesFromCurrentParse(self, currentFsmParse: FsmParse, fsmParse: list, maxLengthOrSurfaceForm,
                                     root: TxtWord):
        currentState = currentFsmParse.getFinalSuffix()
        currentSurfaceForm = currentFsmParse.getSurfaceForm()
        if isinstance(maxLengthOrSurfaceForm, int):
            maxLength = maxLengthOrSurfaceForm
            for currentTransition in self.__finiteStateMachine.getTransitions(currentState):
                if currentTransition.transitionPossibleForParse(currentFsmParse) \
                        and (currentSurfaceForm != root.getName()
                             or (currentSurfaceForm == root.getName() and
                                 currentTransition.transitionPossibleWord(root, currentState))):
                    tmp = currentTransition.makeTransition(root, currentSurfaceForm, currentFsmParse.getStartState())
                    if len(tmp) <= maxLength:
                        newFsmParse = copy.deepcopy(currentFsmParse)
                        newFsmParse.addSuffix(currentTransition.toState(), tmp, currentTransition.withName(),
                                              currentTransition.__str__(), currentTransition.toPos())
                        newFsmParse.setAgreement(currentTransition.withName())
                        fsmParse.append(newFsmParse)
        elif isinstance(maxLengthOrSurfaceForm, str):
            surfaceForm = maxLengthOrSurfaceForm
            for currentTransition in self.__finiteStateMachine.getTransitions(currentState):
                if currentTransition.transitionPossibleForString(currentFsmParse.getSurfaceForm(), surfaceForm) and \
                        currentTransition.transitionPossibleForParse(currentFsmParse) and (
                        currentSurfaceForm != root.getName()
                        or (currentSurfaceForm == root.getName() and
                            currentTransition.transitionPossibleForWord(root, currentState))):
                    tmp = currentTransition.makeTransition(root, currentSurfaceForm, currentFsmParse.getStartState())
                    if (len(tmp) < len(surfaceForm) and self.isPossibleSubstring(tmp, surfaceForm, root)) or \
                            (len(tmp) == len(surfaceForm) and (
                                    root.lastIdropsDuringSuffixation() or tmp == surfaceForm)):
                        newFsmParse = copy.deepcopy(currentFsmParse)
                        newFsmParse.addSuffix(currentTransition.toState(), tmp, currentTransition.withName(),
                                              currentTransition.__str__(), currentTransition.toPos())
                        newFsmParse.setAgreement(currentTransition.withName())
                        fsmParse.append(newFsmParse)

    """
    The parseExists method is used to check the existence of the parse.

    PARAMETERS
    ----------
    fsmParse : list
        List of FsmParse
    surfaceForm : str
        String to use during transition.
        
    RETURNS
    -------
    bool
        True when the currentState is end state and input surfaceForm id equal to currentSurfaceForm, otherwise false.
    """

    def parseExists(self, fsmParse: list, surfaceForm: str) -> bool:
        while len(fsmParse) > 0:
            currentFsmParse = fsmParse.pop(0)
            root = currentFsmParse.getWord()
            currentState = currentFsmParse.getFinalSuffix()
            currentSurfaceForm = currentFsmParse.getSurfaceForm()
            if currentState.isEndState() and currentSurfaceForm == surfaceForm:
                return True
            self.addNewParsesFromCurrentParse(currentFsmParse, fsmParse, surfaceForm, root)
        return False

    """
    The parseWordSurfaceForm method is used to parse a given fsmParse. It simply adds new parses to the current parse by
    using addNewParsesFromCurrentParse method.
    The parseWordMaxLength method is used to parse a given fsmParse. It simply adds new parses to the current parse by
    using addNewParsesFromCurrentParse method.

    PARAMETERS
    ----------
    fsmParse : list   
        a list of FsmParse
    maxLength : int
        maximum length of the surfaceform.
    surfaceForm : str
        String to use during transition.
        
    RETURNS
    -------
    list
        Result list which has the currentFsmParse.
    """
    def parseWord(self, fsmParse: list, maxLengthOrSurfaceForm) -> list:
        result = []
        if isinstance(maxLengthOrSurfaceForm, int):
            maxLength = maxLengthOrSurfaceForm
            while len(fsmParse) > 0:
                currentFsmParse = fsmParse.pop(0)
                root = currentFsmParse.getWord()
                currentState = currentFsmParse.getFinalSuffix()
                currentSurfaceForm = currentFsmParse.getSurfaceForm()
                if currentState.isEndState() and len(currentSurfaceForm) <= maxLength:
                    exists = False
                    for i in range(len(result)):
                        if currentFsmParse.suffixList() == result[i].suffixList():
                            exists = True
                            break
                    if not exists:
                        result.append(currentFsmParse)
                        currentFsmParse.constructInflectionalGroups()
                self.addNewParsesFromCurrentParse(currentFsmParse, fsmParse, maxLength, root)
        elif isinstance(maxLengthOrSurfaceForm, str):
            surfaceForm = maxLengthOrSurfaceForm
            while len(fsmParse) > 0:
                currentFsmParse = fsmParse.pop(0)
                root = currentFsmParse.getWord()
                currentState = currentFsmParse.getFinalSuffix()
                currentSurfaceForm = currentFsmParse.getSurfaceForm()
                if currentState.isEndState() and currentSurfaceForm == surfaceForm:
                    exists = False
                    for i in range(len(result)):
                        if currentFsmParse.suffixList() == result[i].suffixList():
                            exists = True
                            break
                    if not exists:
                        result.append(currentFsmParse)
                        currentFsmParse.constructInflectionalGroups()
                self.addNewParsesFromCurrentParse(currentFsmParse, fsmParse, surfaceForm, root)
        return result

    """
    The morphologicalAnalysis with 3 inputs is used to initialize an {@link ArrayList} and add a new FsmParse
    with given root and state.

    PARAMETERS
    ----------
    root : TxtWord       
        TxtWord input.
    surfaceForm : str
        String input to use for parsing.
    state : str      
        String input.
        
    RETURNS
    -------
    list
        parseWord method with newly populated FsmParse ArrayList and input surfaceForm.
    """

    def morphologicalAnalysisRoot(self, surfaceForm: str, root: TxtWord, state=None) -> list:
        if state is None:
            initialFsmParse = []
            self.initializeParseListFromRoot(initialFsmParse, root, self.isProperNoun(surfaceForm))
        else:
            initialFsmParse = [FsmParse(root, self.__finiteStateMachine.getState(state))]
        return self.parseWord(initialFsmParse, surfaceForm)

    """
    The generateAllParses with 2 inputs is used to generate all parses with given root. Then it calls 
    initializeParseListFromRoot method to initialize list with newly created ArrayList, input root, and maximum length.

    PARAMETERS
    ----------
    root : TxtWord       
        TxtWord input.
    maxLength : int
        Maximum length of the surface form.
        
    RETURNS
    -------
    list
        parseWordMaxLength method with newly populated FsmParse ArrayList and maximum length.
    """

    def generateAllParses(self, root: TxtWord, maxLength: int) -> list:
        initialFsmParse = []
        if root.isProperNoun():
            self.initializeParseListFromRoot(initialFsmParse, root, True)
        self.initializeParseListFromRoot(initialFsmParse, root, False)
        return self.parseWord(initialFsmParse, maxLength)

    """
    The analysisExists method checks several cases. If the given surfaceForm is a punctuation or double then it
    returns true. If it is not a root word, then it initializes the parse list and returns the parseExists method with
    this newly initialized list and surfaceForm.

    PARAMETERS
    ----------
    rootWord : TxtWord    
        TxtWord root.
    surfaceForm : str
        String input.
    isProper : bool   
        boolean variable indicates a word is proper or not.
        
    RETURNS
    -------
    bool
        True if surfaceForm is punctuation or double, otherwise returns parseExist method with given surfaceForm.
    """

    def analysisExists(self, rootWord: TxtWord, surfaceForm: str, isProper: bool) -> bool:
        if Word.isPunctuationSymbol(surfaceForm):
            return True
        if self.isDouble(surfaceForm):
            return True
        if rootWord is not None:
            initialFsmParse = []
            self.initializeParseListFromRoot(initialFsmParse, rootWord, isProper)
        else:
            initialFsmParse = self.initializeParseListFromSurfaceForm(surfaceForm, isProper)
        return self.parseExists(initialFsmParse, surfaceForm)

    """
    The analysis method is used by the morphologicalAnalysis method. It gets String surfaceForm as an input and checks
    its type such as punctuation, number or compares with the regex for date, fraction, percent, time, range, hashtag,
    and mail or checks its variable type as integer or double. After finding the right case for given surfaceForm, it 
    calls constructInflectionalGroups method which creates sub-word units.

    PARAMETERS
    ----------
    surfaceForm : str
        String to analyse.
    isProper : bool   
        is used to indicate the proper words.
        
    RETURNS
    -------
    list
        List type initialFsmParse which holds the analyses.
    """

    def analysis(self, surfaceForm: str, isProper: bool) -> list:
        if Word.isPunctuationSymbol(surfaceForm) and surfaceForm != "%":
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("Punctuation", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.isNumber(surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("CardinalRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.patternMatches("(\\d\\d|\\d)/(\\d\\d|\\d)/\\d+", surfaceForm) or \
                self.patternMatches("(\\d\\d|\\d)\\.(\\d\\d|\\d)\\.\\d+", surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("DateRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.patternMatches("\\d+/\\d+", surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("FractionRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            fsmParse = FsmParse(surfaceForm, State("DateRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.patternMatches("\\d+\\\\/\\d+", surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("FractionRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if surfaceForm == "%" or self.patternMatches("%(\\d\\d|\\d)", surfaceForm) or \
                self.patternMatches("%(\\d\\d|\\d)\\.\\d+", surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("PercentRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.patternMatches("(\\d\\d|\\d):(\\d\\d|\\d):(\\d\\d|\\d)", surfaceForm) or \
                self.patternMatches("(\\d\\d|\\d):(\\d\\d|\\d)", surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("TimeRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.patternMatches("\\d+-\\d+", surfaceForm) or \
                self.patternMatches("(\\d\\d|\\d):(\\d\\d|\\d)-(\\d\\d|\\d):(\\d\\d|\\d)", surfaceForm) or \
                self.patternMatches("(\\d\\d|\\d)\\.(\\d\\d|\\d)-(\\d\\d|\\d)\\.(\\d\\d|\\d)", surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("RangeRoot", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if surfaceForm.startswith("#"):
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("Hashtag", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if "@" in surfaceForm:
            initialFsmParse = []
            fsmParse = FsmParse(surfaceForm, State("Email", True, True))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if surfaceForm.endswith(".") and self.isInteger(surfaceForm[:len(surfaceForm) - 1]):
            initialFsmParse = []
            fsmParse = FsmParse(int(surfaceForm[:len(surfaceForm) - 1]),
                                self.__finiteStateMachine.getState("OrdinalRoot"))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.isInteger(surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(int(surfaceForm), self.__finiteStateMachine.getState("CardinalRoot"))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        if self.isDouble(surfaceForm):
            initialFsmParse = []
            fsmParse = FsmParse(float(surfaceForm), self.__finiteStateMachine.getState("RealRoot"))
            fsmParse.constructInflectionalGroups()
            initialFsmParse.append(fsmParse)
            return initialFsmParse
        initialFsmParse = self.initializeParseListFromSurfaceForm(surfaceForm, isProper)
        return self.parseWord(initialFsmParse, surfaceForm)

    def patternMatches(self, expr: str, value: str) -> bool:
        if expr in self.__mostUsedPatterns:
            return self.__mostUsedPatterns[expr].fullmatch(value) is not None
        else:
            compiledExpression = re.compile(expr)
            self.__mostUsedPatterns[expr] = compiledExpression
            return compiledExpression.fullmatch(value) is not None

    """
    The isProperNoun method takes surfaceForm String as input and checks its each char whether they are in the range
    of letters between A to Z or one of the Turkish letters such as İ, Ü, Ğ, Ş, Ç, and Ö.

    PARAMETERS
    ----------
    surfaceForm : str
        String to check for proper noun.
        
    RETURNS
    -------
    bool
        False if surfaceForm is null or length of 0, return true if it is a letter.
    """

    def isProperNoun(self, surfaceForm: str) -> bool:
        if surfaceForm is None or len(surfaceForm) == 0:
            return False
        return "A" <= surfaceForm[0] <= "Z" or surfaceForm[0] == "Ç" or surfaceForm[0] == "Ö" or surfaceForm[0] == "Ğ" \
               or surfaceForm[0] == "Ü" or surfaceForm[0] == "Ş" or surfaceForm[0] == "İ"

    """
    The morphologicalAnalysis method is used to analyse a FsmParseList by comparing with the regex.
    It creates a list fsmParse to hold the result of the analysis method. For each surfaceForm input,
    it gets a substring and considers it as a possibleRoot. Then compares with the regex.

    If the surfaceForm input string matches with Turkish chars like Ç, Ş, İ, Ü, Ö, it adds the surfaceForm to Trie with 
    IS_OA tag.
    If the possibleRoot contains /, then it is added to the Trie with IS_KESIR tag.
    If the possibleRoot contains \\d\\d|\\d)/(\\d\\d|\\d)/\\d+, then it is added to the Trie with IS_DATE tag.
    If the possibleRoot contains \\d\\d|\\d, then it is added to the Trie with IS_PERCENT tag.
    If the possibleRoot contains \\d\\d|\\d):(\\d\\d|\\d):(\\d\\d|\\d), then it is added to the Trie with IS_ZAMAN tag.
    If the possibleRoot contains \\d+-\\d+, then it is added to the Trie with IS_RANGE tag.
    If the possibleRoot is an Integer, then it is added to the Trie with IS_SAYI tag.
    If the possibleRoot is a Double, then it is added to the Trie with IS_REELSAYI tag.

    PARAMETERS
    ----------
    sentenceOrSurfaceForm : str
        String or Sentence to analyse.

    RETURNS
    -------
    list
        list which folds the analysis
    FsmParseList
        fsmParseList which holds the analysis.
    """

    def morphologicalAnalysis(self, sentenceOrSurfaceForm):
        if isinstance(sentenceOrSurfaceForm, Sentence):
            sentence = sentenceOrSurfaceForm
            result = []
            for i in range(sentence.wordCount()):
                originalForm = sentence.getWord(i).getName()
                spellCorrectedForm = self.__dictionary.getCorrectForm(originalForm)
                if len(spellCorrectedForm) == 0:
                    spellCorrectedForm = originalForm
                wordFsmParseList = self.morphologicalAnalysis(spellCorrectedForm)
                result.append(wordFsmParseList)
            return result
        elif isinstance(sentenceOrSurfaceForm, str):
            surfaceForm = sentenceOrSurfaceForm
            if self.__cache.contains(surfaceForm):
                return self.__cache.get(surfaceForm)
            if self.patternMatches("(\\w|Ç|Ş|İ|Ü|Ö)\\.", surfaceForm):
                self.__dictionaryTrie.addWord(self.toLower(surfaceForm), TxtWord(self.toLower(surfaceForm), "IS_OA"))
            defaultFsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
            if len(defaultFsmParse) > 0:
                fsmParseList = FsmParseList(defaultFsmParse)
                self.__cache.add(surfaceForm, fsmParseList)
                return fsmParseList
            fsmParse = []
            if "'" in surfaceForm:
                possibleRoot = surfaceForm[:surfaceForm.index('\'')]
                if len(possibleRoot) > 0:
                    if "/" in possibleRoot or "\\/" in possibleRoot:
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_KESIR"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif self.patternMatches("(\\d\\d|\\d)/(\\d\\d|\\d)/\\d+", possibleRoot) or \
                            self.patternMatches("(\\d\\d|\\d)\\.(\\d\\d|\\d)\\.\\d+", possibleRoot):
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_DATE"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif self.patternMatches("\\d+/\\d+", possibleRoot):
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_KESIR"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif self.patternMatches("%(\\d\\d|\\d)", possibleRoot) or \
                            self.patternMatches("%(\\d\\d|\\d)\\.\\d+", possibleRoot):
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_PERCENT"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif self.patternMatches("(\\d\\d|\\d):(\\d\\d|\\d):(\\d\\d|\\d)", possibleRoot) or \
                            self.patternMatches("(\\d\\d|\\d):(\\d\\d|\\d)", possibleRoot):
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_ZAMAN"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif self.patternMatches("\\d+-\\d+", possibleRoot) or \
                            self.patternMatches("(\\d\\d|\\d):(\\d\\d|\\d)-(\\d\\d|\\d):(\\d\\d|\\d)", possibleRoot) or\
                            self.patternMatches("(\\d\\d|\\d)\\.(\\d\\d|\\d)-(\\d\\d|\\d)\\.(\\d\\d|\\d)",
                                                possibleRoot):
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_RANGE"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif self.isInteger(possibleRoot):
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_SAYI"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif self.isDouble(possibleRoot):
                        self.__dictionaryTrie.addWord(possibleRoot, TxtWord(possibleRoot, "IS_REELSAYI"))
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                    elif Word.isCapital(possibleRoot):
                        newWord = None
                        word = self.__dictionary.getWord(self.toLower(possibleRoot))
                        if word is not None and isinstance(word, TxtWord):
                            word.addFlag("IS_OA")
                        else:
                            newWord = TxtWord(self.toLower(possibleRoot), "IS_OA")
                            self.__dictionaryTrie.addWord(self.toLower(possibleRoot), newWord)
                        fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
                        if len(fsmParse) == 0 and newWord is not None:
                            newWord.addFlag("IS_KIS")
                            fsmParse = self.analysis(self.toLower(surfaceForm), self.isProperNoun(surfaceForm))
            fsmParseList = FsmParseList(fsmParse)
            self.__cache.add(surfaceForm, fsmParseList)
            return fsmParseList

    """
    The robustMorphologicalAnalysis is used to analyse surfaceForm String. First it gets the currentParse of the 
    surfaceForm then, if the size of the currentParse is 0, and given surfaceForm is a proper noun, it adds the 
    surfaceForm whose state name is ProperRoot to an list, of it is not a proper noon, it adds the surfaceForm
    whose state name is NominalRoot to the list.
    The robustMorphologicalAnalysis method takes just one argument as an input. It gets the name of the words from
    input sentence then calls robustMorphologicalAnalysis with surfaceForm.

    PARAMETERS
    ----------
    sentence : Sentence
        Sentence type input used to get surfaceForm.
    surfaceForm : str
        String to analyse.
        
    RETURNS
    -------
    list
        FsmParseList array which holds the result of the analysis.
    FsmParseList
        FsmParseList type currentParse which holds morphological analysis of the surfaceForm.
    """

    def robustMorphologicalAnalysis(self, sentenceOrSurfaceForm):
        if isinstance(sentenceOrSurfaceForm, Sentence):
            sentence = sentenceOrSurfaceForm
            result = []
            for i in range(sentence.wordCount()):
                originalForm = sentence.getWord(i).getName()
                spellCorrectedForm = self.__dictionary.getCorrectForm(originalForm)
                if len(spellCorrectedForm) == 0:
                    spellCorrectedForm = originalForm
                wordFsmParseList = self.robustMorphologicalAnalysis(spellCorrectedForm)
                result.append(wordFsmParseList)
            return result
        elif isinstance(sentenceOrSurfaceForm, str):
            surfaceForm = sentenceOrSurfaceForm
            if surfaceForm is None or len(surfaceForm) == 0:
                return FsmParseList([])
            currentParse = self.morphologicalAnalysis(surfaceForm)
            if currentParse.size() == 0:
                fsmParse = []
                if self.isProperNoun(surfaceForm):
                    fsmParse.append(FsmParse(surfaceForm, self.__finiteStateMachine.getState("ProperRoot")))
                    return FsmParseList(self.parseWord(fsmParse, surfaceForm))
                else:
                    fsmParse.append(FsmParse(surfaceForm, self.__finiteStateMachine.getState("NominalRoot")))
                    return FsmParseList(self.parseWord(fsmParse, surfaceForm))
            else:
                return currentParse

    """
    The isInteger method compares input surfaceForm with given regex and returns the result.
    Supports positive integer checks only.

    PARAMETERS
    ----------
    surfaceForm : str
        String to check.
        
    RETURNS
    -------
    bool
        True if surfaceForm matches with the regex.
    """

    def isInteger(self, surfaceForm: str) -> bool:
        if not self.patternMatches("\\+?\\d+", surfaceForm):
            return False
        length = len(surfaceForm)
        if length < 10:
            return True
        elif length > 10:
            return False
        else:
            return surfaceForm <= "2147483647"

    """
    The isDouble method compares input surfaceForm with given regex and returns the result.

    PARAMETERS
    ----------
    surfaceForm : str
        String to check.
        
    RETURNS
    -------
    bool
        True if surfaceForm matches with the regex.
    """

    def isDouble(self, surfaceForm: str) -> bool:
        return self.patternMatches("\\+?(\\d+)?\\.\\d*", surfaceForm)

    """
    The isNumber method compares input surfaceForm with the array of written numbers and returns the result.

    PARAMETERS
    ----------
    surfaceForm : str
        String to check.
        
    RETURNS
    -------
    bool
        True if surfaceForm matches with the regex.
    """

    def isNumber(self, surfaceForm: str) -> bool:
        numbers = ["bir", "iki", "üç", "dört", "beş", "altı", "yedi", "sekiz", "dokuz",
                   "on", "yirmi", "otuz", "kırk", "elli", "altmış", "yetmiş", "seksen", "doksan",
                   "yüz", "bin", "milyon", "milyar", "trilyon", "katrilyon"]
        word = surfaceForm
        count = 0
        while len(word) > 0:
            found = False
            for number in numbers:
                if word.startswith(number):
                    found = True
                    count = count + 1
                    word = word[len(number):]
                    break
            if not found:
                break
        return len(word) == 0 and count > 1

    def toLower(self, surfaceForm: str) -> str:
        if "I" in surfaceForm or "İ" in surfaceForm:
            result = ""
            for i in range(len(surfaceForm)):
                if surfaceForm[i] == "I":
                    result = result + "ı"
                elif surfaceForm[i] == "İ":
                    result = result + "i"
                else:
                    result = result + surfaceForm[i].lower()
            return result
        else:
            return surfaceForm.lower()

    """
    The morphologicalAnalysisExists method calls analysisExists to check the existence of the analysis with given
    root and surfaceForm.

    PARAMETERS
    ----------
    surfaceForm : str
        String to check.
    rootWord : TxtWord   
        TxtWord input root.
        
    RETURNS
    -------
    bool
        True an analysis exists, otherwise return False.
    """

    def morphologicalAnalysisExists(self, rootWord: TxtWord, surfaceForm: str) -> bool:
        return self.analysisExists(rootWord, self.toLower(surfaceForm), True)



from MorphologicalAnalysis.MorphologicalTag import MorphologicalTag


class InflectionalGroup:
    __IG: list
    tags = ["NOUN", "ADV", "ADJ", "VERB", "A1SG",
            "A2SG", "A3SG", "A1PL", "A2PL", "A3PL",
            "P1SG", "P2SG", "P3SG", "P1PL", "P2PL",
            "P3PL", "PROP", "PNON", "NOM", "WITH",
            "WITHOUT", "ACC", "DAT", "GEN", "ABL",
            "ZERO", "ABLE", "NEG", "PAST",
            "CONJ", "DET", "DUP", "INTERJ", "NUM",
            "POSTP", "PUNC", "QUES", "AGT", "BYDOINGSO",
            "CARD", "CAUS", "DEMONSP", "DISTRIB", "FITFOR",
            "FUTPART", "INF", "NESS", "ORD", "PASS",
            "PASTPART", "PRESPART", "QUESP", "QUANTP", "RANGE",
            "RATIO", "REAL", "RECIP", "REFLEX", "REFLEXP",
            "TIME", "WHEN", "WHILE", "WITHOUTHAVINGDONESO", "PCABL",
            "PCACC", "PCDAT", "PCGEN", "PCINS", "PCNOM",
            "ACQUIRE", "ACTOF", "AFTERDOINGSO", "ALMOST", "AS",
            "ASIF", "BECOME", "EVERSINCE", "FEELLIKE", "HASTILY",
            "INBETWEEN", "JUSTLIKE", "LY", "NOTABLESTATE", "RELATED",
            "REPEAT", "SINCE", "SINCEDOINGSO", "START", "STAY",
            "EQU", "INS", "AOR", "DESR", "FUT",
            "IMP", "NARR", "NECES", "OPT", "PAST",
            "PRES", "PROG1", "PROG2", "COND", "COP",
            "POS", "PRON", "LOC", "REL", "DEMONS",
            "INF2", "INF3", "BSTAG", "ESTAG", "BTTAG",
            "ETTAG", "BDTAG", "EDTAG", "INF1", "ASLONGAS",
            "DIST", "ADAMANTLY", "PERCENT", "WITHOUTBEINGABLETOHAVEDONESO", "DIM",
            "PERS", "FRACTION", "HASHTAG", "EMAIL", "DATE"]
    morphotags = [MorphologicalTag.NOUN, MorphologicalTag.ADVERB, MorphologicalTag.ADJECTIVE,
                  MorphologicalTag.VERB, MorphologicalTag.A1SG, MorphologicalTag.A2SG, MorphologicalTag.A3SG,
                  MorphologicalTag.A1PL,
                  MorphologicalTag.A2PL, MorphologicalTag.A3PL, MorphologicalTag.P1SG, MorphologicalTag.P2SG,
                  MorphologicalTag.P3SG, MorphologicalTag.P1PL,
                  MorphologicalTag.P2PL, MorphologicalTag.P3PL, MorphologicalTag.PROPERNOUN, MorphologicalTag.PNON,
                  MorphologicalTag.NOMINATIVE,
                  MorphologicalTag.WITH, MorphologicalTag.WITHOUT, MorphologicalTag.ACCUSATIVE, MorphologicalTag.DATIVE,
                  MorphologicalTag.GENITIVE,
                  MorphologicalTag.ABLATIVE, MorphologicalTag.ZERO, MorphologicalTag.ABLE, MorphologicalTag.NEGATIVE,
                  MorphologicalTag.PASTTENSE,
                  MorphologicalTag.CONJUNCTION, MorphologicalTag.DETERMINER, MorphologicalTag.DUPLICATION,
                  MorphologicalTag.INTERJECTION, MorphologicalTag.NUMBER,
                  MorphologicalTag.POSTPOSITION, MorphologicalTag.PUNCTUATION, MorphologicalTag.QUESTION,
                  MorphologicalTag.AGENT, MorphologicalTag.BYDOINGSO,
                  MorphologicalTag.CARDINAL, MorphologicalTag.CAUSATIVE, MorphologicalTag.DEMONSTRATIVEPRONOUN,
                  MorphologicalTag.DISTRIBUTIVE, MorphologicalTag.FITFOR,
                  MorphologicalTag.FUTUREPARTICIPLE, MorphologicalTag.INFINITIVE, MorphologicalTag.NESS,
                  MorphologicalTag.ORDINAL, MorphologicalTag.PASSIVE,
                  MorphologicalTag.PASTPARTICIPLE, MorphologicalTag.PRESENTPARTICIPLE, MorphologicalTag.QUESTIONPRONOUN,
                  MorphologicalTag.QUANTITATIVEPRONOUN, MorphologicalTag.RANGE,
                  MorphologicalTag.RATIO, MorphologicalTag.REAL, MorphologicalTag.RECIPROCAL,
                  MorphologicalTag.REFLEXIVE, MorphologicalTag.REFLEXIVEPRONOUN,
                  MorphologicalTag.TIME, MorphologicalTag.WHEN, MorphologicalTag.WHILE,
                  MorphologicalTag.WITHOUTHAVINGDONESO, MorphologicalTag.PCABLATIVE,
                  MorphologicalTag.PCACCUSATIVE, MorphologicalTag.PCDATIVE, MorphologicalTag.PCGENITIVE,
                  MorphologicalTag.PCINSTRUMENTAL, MorphologicalTag.PCNOMINATIVE,
                  MorphologicalTag.ACQUIRE, MorphologicalTag.ACTOF, MorphologicalTag.AFTERDOINGSO,
                  MorphologicalTag.ALMOST, MorphologicalTag.AS,
                  MorphologicalTag.ASIF, MorphologicalTag.BECOME, MorphologicalTag.EVERSINCE, MorphologicalTag.FEELLIKE,
                  MorphologicalTag.HASTILY,
                  MorphologicalTag.INBETWEEN, MorphologicalTag.JUSTLIKE, MorphologicalTag.LY,
                  MorphologicalTag.NOTABLESTATE, MorphologicalTag.RELATED,
                  MorphologicalTag.REPEAT, MorphologicalTag.SINCE, MorphologicalTag.SINCEDOINGSO,
                  MorphologicalTag.START, MorphologicalTag.STAY,
                  MorphologicalTag.EQUATIVE, MorphologicalTag.INSTRUMENTAL, MorphologicalTag.AORIST,
                  MorphologicalTag.DESIRE, MorphologicalTag.FUTURE,
                  MorphologicalTag.IMPERATIVE, MorphologicalTag.NARRATIVE, MorphologicalTag.NECESSITY,
                  MorphologicalTag.OPTATIVE, MorphologicalTag.PAST,
                  MorphologicalTag.PRESENT, MorphologicalTag.PROGRESSIVE1, MorphologicalTag.PROGRESSIVE2,
                  MorphologicalTag.CONDITIONAL, MorphologicalTag.COPULA,
                  MorphologicalTag.POSITIVE, MorphologicalTag.PRONOUN, MorphologicalTag.LOCATIVE,
                  MorphologicalTag.RELATIVE, MorphologicalTag.DEMONSTRATIVE,
                  MorphologicalTag.INFINITIVE2, MorphologicalTag.INFINITIVE3, MorphologicalTag.BEGINNINGOFSENTENCE,
                  MorphologicalTag.ENDOFSENTENCE, MorphologicalTag.BEGINNINGOFTITLE,
                  MorphologicalTag.ENDOFTITLE, MorphologicalTag.BEGINNINGOFDOCUMENT, MorphologicalTag.ENDOFDOCUMENT,
                  MorphologicalTag.INFINITIVE, MorphologicalTag.ASLONGAS,
                  MorphologicalTag.DISTRIBUTIVE, MorphologicalTag.ADAMANTLY, MorphologicalTag.PERCENT,
                  MorphologicalTag.WITHOUTBEINGABLETOHAVEDONESO, MorphologicalTag.DIMENSION,
                  MorphologicalTag.PERSONALPRONOUN, MorphologicalTag.FRACTION, MorphologicalTag.HASHTAG,
                  MorphologicalTag.EMAIL, MorphologicalTag.DATE]

    """
    A constructor of InflectionalGroup class which initializes the IG list by parsing given input
    String IG by + and calling the getMorphologicalTag method with these substrings. If getMorphologicalTag method 
    returns a tag, it adds this tag to the IG list.

    PARAMETERS
    ----------
    IG : str
        String input.
    """
    def __init__(self, IG: str):
        self.__IG = []
        st = IG
        while "+" in st:
            morphologicalTag = st[:st.index("+")]
            tag = InflectionalGroup.getMorphologicalTag(morphologicalTag)
            if tag is not None:
                self.__IG.append(tag)
            else:
                print("Morphological Tag " + morphologicalTag + " not found")
            st = st[st.index("+") + 1:]
        morphologicalTag = st
        tag = InflectionalGroup.getMorphologicalTag(morphologicalTag)
        if tag is not None:
            self.__IG.append(tag)
        else:
            print("Morphological Tag " + morphologicalTag + " not found")

    """
    The getMorphologicalTag method takes a String tag as an input and if the input matches with one of the elements of
    tags array, it then gets the morphoTags of this tag and returns it.

    PARAMETERS
    ----------
    tag : str
        String to get morphoTags from.
        
    RETURNS
    -------
    MorphologicalTag
        morphoTags if found, None otherwise.
    """
    @staticmethod
    def getMorphologicalTag(tag: str) -> MorphologicalTag:
        for j in range(len(InflectionalGroup.tags)):
            if tag == InflectionalGroup.tags[j]:
                return InflectionalGroup.morphotags[j]
        return None

    """
    The getTag method takes a MorphologicalTag type tag as an input and returns its corresponding tag from tags array.

    PARAMETERS
    ----------
    tag : MorphologicalTag
        MorphologicalTag type input to find tag from.
        
    RETURNS
    -------
    str
        Tag if found, None otherwise.
    """
    @staticmethod
    def getTagString(tag: MorphologicalTag) -> str:
        for j in range(len(InflectionalGroup.morphotags)):
            if tag == InflectionalGroup.morphotags[j]:
                return InflectionalGroup.tags[j]
        return None

    """
    Another getTag method which takes index as an input and returns the corresponding tag from IG {@link ArrayList}.

    PARAMETERS
    ----------
    index : int
        index to get tag.
        
    RETURNS
    -------
    MorphologicalTag
        tag at input index.
    """
    def getTag(self, index: int) -> MorphologicalTag:
        return self.__IG[index]

    """
    The size method returns the size of the IG list.

    RETURNS
    -------
    int
        the size of the IG list.
    """
    def size(self) -> int:
        return len(self.__IG)

    """
    Overridden toString method to return resulting tags in IG list.

    RETURNS
    -------
    str
        String result.
    """
    def __str__(self) -> str:
        result = InflectionalGroup.getTagString(self.__IG[0])
        for i in range(1, len(self.__IG)):
            result = result + "+" + InflectionalGroup.getTagString(self.__IG[i])
        return result

    """
    The containsCase method loops through the tags in IG list and finds out the tags of the NOMINATIVE,
    ACCUSATIVE, DATIVE, LOCATIVE or ABLATIVE cases.

    RETURNS
    -------
    MorphologicalTag
        tag which holds the condition.
    """
    def containsCase(self) -> MorphologicalTag:
        for tag in self.__IG:
            if tag == MorphologicalTag.NOMINATIVE or tag == MorphologicalTag.ACCUSATIVE or \
                    tag == MorphologicalTag.DATIVE or tag == MorphologicalTag.LOCATIVE or \
                    tag == MorphologicalTag.ABLATIVE:
                return tag
        return None

    """
    The containsPlural method loops through the tags in IG list and checks whether the tags are from
    the agreement plural or possessive plural, i.e A1PL, A2PL, A3PL, P1PL, P2PL and P3PL.

    RETURNS
    -------
    bool
        True if the tag is plural, False otherwise.
    """
    def containsPlural(self) -> bool:
        for tag in self.__IG:
            if tag == MorphologicalTag.A1PL or tag == MorphologicalTag.A2PL or tag == MorphologicalTag.A3PL or \
                    tag == MorphologicalTag.P1PL or tag == MorphologicalTag.P2PL or tag == MorphologicalTag.P3PL:
                return True
        return False

    """
    The containsTag method takes a MorphologicalTag type tag as an input and loops through the tags in
    IG list and returns true if the input matches with on of the tags in the IG.

    PARAMETERS
    ----------
    tag : MorphologicalTag
        MorphologicalTag type input to search for.
        
    RETURNS
    -------
    bool
        True if tag matches with the tag in IG, False otherwise.
    """
    def containsTag(self, tag: MorphologicalTag) -> bool:
        for currentTag in self.__IG:
            if tag == currentTag:
                return True
        return False

    """
    The containsPossessive method loops through the tags in IG list and returns true if the tag in IG is
    one of the possessives: P1PL, P1SG, P2PL, P2SG, P3PL AND P3SG.

    RETURNS
    -------
    bool
        True if it contains possessive tag, False otherwise.
    """
    def containsPossessive(self) -> bool:
        for tag in self.__IG:
            if tag == MorphologicalTag.P1SG or tag == MorphologicalTag.P1PL or tag == MorphologicalTag.P2SG or \
                    tag == MorphologicalTag.P2PL or tag == MorphologicalTag.P3SG or tag == MorphologicalTag.P3PL:
                return True
        return False
